import pathlib
import logging
import panflute as pf

from pandoc_filter.utils import TracingLogger
from pandoc_filter.utils import OssHelper
from pandoc_filter.md2md_filters import math_filter,figure_filter,footnote_filter,internal_link_filter


pathlib.Path("./logs").mkdir(parents=True, exist_ok=True)
tracing_logger = TracingLogger(name="./logs/pf_log",level=logging.INFO)

def _check_file_path(file_path:str)->pathlib.Path:
    file_path:pathlib.Path = pathlib.Path(file_path)
    assert file_path.exists()
    assert file_path.is_file()
    return file_path
    
def test_md2md_footnote_filter():
    file_path =_check_file_path("./resources/test_md2md_footnote.md")
    with open(file_path,'r',encoding='utf-8') as f:
        markdown_content = f.read()
    output_path = pathlib.Path(f"./temp/{file_path.name}")
   
    doc = pf.convert_text(markdown_content,input_format='markdown',output_format='panflute',standalone=True)
    doc = pf.run_filter(action=footnote_filter,doc=doc,tracing_logger=tracing_logger)
    with open(output_path, "w", encoding="utf-8") as f:
        f.write(pf.convert_text(doc,input_format='panflute',output_format='gfm',standalone=True))

def test_md2md_internal_link_filter():
    file_path = _check_file_path("./resources/test_md2md_internal_link.md")
    with open(file_path,'r',encoding='utf-8') as f:
        markdown_content = f.read()
    output_path = pathlib.Path(f"./temp/{file_path.name}")

    doc = pf.convert_text(markdown_content,input_format='markdown',output_format='panflute',standalone=True)
    doc = pf.run_filter(action=internal_link_filter,doc=doc,tracing_logger=tracing_logger)
    with open(output_path, "w", encoding="utf-8") as f:
        f.write(pf.convert_text(doc,input_format='panflute',output_format='gfm',standalone=True))

def test_md2md_math_filter():
    file_path = _check_file_path("./resources/test_md2md_math.md")
    with open(file_path,'r',encoding='utf-8') as f:
        markdown_content = f.read()
    output_path = pathlib.Path(f"./temp/{file_path.name}")
    doc = pf.convert_text(markdown_content,input_format='markdown',output_format='panflute',standalone=True)
    doc = pf.run_filter(action=math_filter,doc=doc,tracing_logger=tracing_logger)
    with open(output_path, "w", encoding="utf-8") as f:
        f.write(pf.convert_text(doc,input_format='panflute',output_format='gfm',standalone=True))

def test_md2md_figure_filter():
    file_path = _check_file_path("./resources/test_md2md_figure.md")
    with open(file_path,'r',encoding='utf-8') as f:
        markdown_content = f.read()
    import os
    oss_endpoint_name = os.environ['OSS_ENDPOINT_NAME']
    oss_bucket_name = os.environ['OSS_BUCKET_NAME']
    assert os.environ['OSS_ACCESS_KEY_ID']
    assert os.environ['OSS_ACCESS_KEY_SECRET']
    oss_helper = OssHelper(oss_endpoint_name,oss_bucket_name)
    output_path = pathlib.Path(f"./temp/{file_path.name}")
    doc = pf.convert_text(markdown_content,input_format='markdown',output_format='panflute',standalone=True)
    doc.doc_path = file_path
    doc = pf.run_filter(action=figure_filter,doc=doc,tracing_logger=tracing_logger,oss_helper=oss_helper)
    with open(output_path, "w", encoding="utf-8") as f:
        f.write(pf.convert_text(doc,input_format='panflute',output_format='gfm',standalone=True))
      



import logging
import re
import typeguard
import urllib.parse
import pathlib
import panflute as pf
from .utils.logging_helper import TracingLogger
from .utils.oss_helper import OssHelper
from .utils import html_helper
r"""Defination
In Markdown:
    anchors:
        headings anchors: `## aaa`
        raw-HTML anchors: `<a id="aaa"></a>`
    links:
        internal links:
            md internal links: `[bbb](#aaa)`
            raw-HTML internal links: `<a href="#aaa">bbb</a>`
        ...
    ...
"""
def _decode_internal_link_url(url:str)->str:
    r"""When converting markdown to any type via pandoc, md internal links' URLs may be automatically URL-encoded before any filter works.
    The encoding is done by default and may not be avoided.
    This function is used to decode the URL.
    """
    decoded_url = urllib.parse.unquote(url.lstrip('#'))
    header_mimic = pf.convert_text(f"# {decoded_url}",input_format='markdown',output_format='gfm',standalone=True)
    return f"#{header_mimic.lstrip('# ')}"
    
@typeguard.typechecked
def internal_link_filter(elem:pf.Element, doc:pf.Doc,**kwargs)->None: # Do not modify.
    r"""Follow the general procedure of [Panflute](http://scorreia.com/software/panflute/)
    A filter to normalize internal links when converting markdown to markdown.
    """
    tracing_logger:TracingLogger = kwargs['tracing_logger']
    if isinstance(elem, pf.Link) and elem.url.startswith('#'):
        tracing_logger.mark(elem)       
        elem.url = _decode_internal_link_url(elem.url)
        tracing_logger.check_and_log('anchor_links',elem)
    elif isinstance(elem, pf.RawInline) and elem.format == 'html' and (old_href:=html_helper.get_href(elem.text)) and old_href.startswith('#'):
        tracing_logger.mark(elem)
        elem.text = html_helper.sub_href(elem.text,_decode_internal_link_url(old_href))
        tracing_logger.check_and_log('raw_anchor_links',elem)


@typeguard.typechecked
def math_filter(elem:pf.Element,doc:pf.Doc,**kwargs)->None: # Modify In Place
    r"""Follow the general procedure of [Panflute](http://scorreia.com/software/panflute/)
    A filter to process math formula when converting markdown to markdown.
    To realize:
        - Adapt AMS rule for math formula
            - Auto numbering markdown formulations within \begin{equation} \end{equation}, as in Typora
        - Allow multiple tags, but only take the first one.
        - Allow multiple labels, but only take the first one.
        
    To make equations recognized correctly, content like the following irregular syntax should to be avoided:
    
        ```markdown
        \begin{equation}\tag{33124124}
        e=mc^2 \\
        e=mc^2 \\

        \begin{aligned}
        \\
        \\
        \\
        e=mc^2 \\
        e=mc^2 \\
        \end{aligned}
        \end{equation}
        ```
    """
    tracing_logger:TracingLogger = kwargs['tracing_logger']
    if not(hasattr(doc,'equations_count') and isinstance(doc.equations_count,int) and (doc.equations_count >= 0)):
        doc.equations_count = 0
    if not(hasattr(doc,'have_math') and isinstance(doc.have_math,bool)):
        doc.have_math = False
    if isinstance(elem, pf.elements.Math):
        if not doc.have_math: # lazy modification
            doc.have_math = True
        if elem.format == "DisplayMath":
            tracing_logger.mark(elem)
            text = elem.text
            # delete all labels and tags but record the first one
            first_label = ''
            if matched:= re.search(r"(\\label{[^{}]*})",text): 
                first_label = matched.group(0)
                text = re.sub(r"(\\label{[^{}]*})",'',text)
            first_tag = ''
            if matched:= re.search(r"(\\tag{[^{}]*})",text): 
                first_tag = matched.group(0)
                text = re.sub(r"(\\tag{[^{}]*})",'',text)
            
            if (re.search(r"^\s*\\begin{equation}",text) and re.search(r"\\end{equation}\s*$",text)):
                text = re.sub(r"^\s*\\begin{equation}",'',text)
                text = re.sub(r"\\end{equation}\s*$",'',text)
                
                if first_tag != '':
                    text = f"\\begin{{equation}}{first_label}{first_tag}\n{text.strip(" \n")}\n\\end{{equation}}"
                else:
                    doc.equations_count += 1
                    text = f"\\begin{{equation}}{first_label}\\tag{{{doc.equations_count}}}\n{text.strip(" \n")}\n\\end{{equation}}"
            else:
                text = f"{text}\n{first_label}{first_tag}"
            elem.text = f"\n{text.strip(" \n")}\n"
            tracing_logger.check_and_log('equation',elem)

@typeguard.typechecked     
def figure_filter(elem:pf.Element,doc:pf.Doc,**kwargs)->None: # Modify In Place
    r"""Follow the general procedure of [Panflute](http://scorreia.com/software/panflute/)
    A filter to process images of local pictures when converting markdown to markdown.
    Manager local pictures, sync them to Aliyun OSS, and replace the original src with the new one.
    NOTE
        The `doc.doc_path` should be set before calling this filter.
        The `doc.doc_path` should be a pathlib.Path object.
        The local pictures should be within the same directory as the doc file.
    """
    oss_helper:OssHelper = kwargs['oss_helper']
    tracing_logger:TracingLogger = kwargs['tracing_logger']
    if not(hasattr(doc,'doc_path') and isinstance(doc.doc_path,pathlib.Path)):
        tracing_logger.logger.warning("doc.doc_path is not set or is not a pathlib.Path object.")
        return None    
    if isinstance(elem, pf.Image) and (old_src:=str(elem.url)).startswith('.'): # reletive path
        new_src = oss_helper.maybe_upload_file_and_get_src(doc.doc_path.parent/old_src)
        tracing_logger.mark(elem)
        elem.url = new_src
        tracing_logger.check_and_log('image',elem)
    elif isinstance(elem, pf.RawInline) and elem.format == 'html' and (old_src:=html_helper.get_src(elem.text)) and old_src.startswith('.'): # reletive path
            new_src = oss_helper.maybe_upload_file_and_get_src(doc.doc_path.parent/old_src)
            tracing_logger.mark(elem)
            elem.text = html_helper.sub_src(elem.text,new_src)
            tracing_logger.check_and_log('raw_html_img',elem)

@typeguard.typechecked
def footnote_filter(elem:pf.Element,doc:pf.Doc,**kwargs)->pf.Note|None: # Repleace
    r"""Follow the general procedure of [Panflute](http://scorreia.com/software/panflute/)
    A filter to process footnotes. Remove `\n` in the footnote content.
    """
    tracing_logger:TracingLogger = kwargs['tracing_logger']
    if isinstance(elem, pf.Note):
        tracing_logger.mark(elem)
        elem = pf.Note(pf.Para(pf.Str(pf.stringify(elem.content).strip(" \n"))))
        tracing_logger.check_and_log('footnote',elem)
        return elem




#!/usr/bin/env python
import re
import sys
import os
import logging

import pathlib
import functools
import typeguard
import requests
from alive_progress import alive_bar

from ..logging_helper import logger_factory

script_file_name = os.path.basename(__file__)

logger = logger_factory(os.path.splitext(script_file_name)[0],logging.WARNING)

def get_normalized_src_alt_filepath(src:str,alt:str,img_save_dir:pathlib.Path):
    
    img_save_dir.mkdir(parents=True,exist_ok=True)
            
    file_extension = src.split('.')[-1]
    
    file_name = re.sub(r"[^\w\u4e00-\u9fa5]+", "-", alt) 
    file_name = file_name.strip('-')
    
    file_path = img_save_dir/f'{file_name}.{file_extension}'
    
    new_src = f"./{file_path.relative_to(img_save_dir.parent).as_posix()}"
    new_alt = file_name
    logger.info(f'new alt: {new_alt}')
    logger.info(f'new src: {new_src}')
    return new_src,new_alt,file_path   
    
    
def sub_html_img_pattern(match:re.Match,img_save_dir:str):
    group_dict = match.groupdict()
    src = group_dict['src']
    alt = group_dict['alt']
    style = group_dict['style']
    img_save_dir:pathlib.Path = pathlib.Path(img_save_dir)
    try:
        response = requests.get(src)
        if response.status_code == 200:
            # save local image
            # normlize file name -> normlize src and alt
            new_src,new_alt,file_path = get_normalized_src_alt_filepath(src,alt,img_save_dir)
            with open(file_path, 'wb') as img_file:
                img_file.write(response.content)
            return f'<img src="{new_src}" alt="{new_alt}" style="{style}"/>'
    except requests.exceptions.MissingSchema as e:
        # save local image
        # normlize file name -> normlize src and alt
        new_src,new_alt,file_path = get_normalized_src_alt_filepath(src,alt,img_save_dir)
        with open(img_save_dir.parent/src, 'rb') as img_file:
            content = img_file.read()
        with open(file_path, 'wb') as img_file:
            img_file.write(content)
        return f'<img src="{new_src}" alt="{new_alt}" style="{style}"/>'
    except Exception as e:
        logger.error(e)
        return f'<img src="{src}" alt="{alt}" style="{style}"/>'

def sub_markdown_img_pattern(match:re.Match,img_save_dir:str):
    group_dict = match.groupdict()
    src = group_dict['src']
    alt = group_dict['alt']
    img_save_dir:pathlib.Path = pathlib.Path(img_save_dir)
    try:
        response = requests.get(src)
        if response.status_code == 200:
            # save network image
            # normlize file name -> normlize src and alt
            new_src,new_alt,file_path = get_normalized_src_alt_filepath(src,alt,img_save_dir)
            with open(file_path, 'wb') as img_file:
                img_file.write(response.content)
            return f'![{new_alt}]({new_src})'
    except requests.exceptions.MissingSchema as e:
        # save local image
        # normlize file name -> normlize src and alt
        new_src,new_alt,file_path = get_normalized_src_alt_filepath(src,alt,img_save_dir)
        with open(img_save_dir.parent/src, 'rb') as img_file:
            content = img_file.read()
        with open(file_path, 'wb') as img_file:
            img_file.write(content)
        return f'![{new_alt}]({new_src})'
    except Exception as e:
        logger.error(e)
        return f'![{alt}]({src})'
    
def main(directory:str):
    """
    Localize all images of all markdown files in the directory. Specifically, the procedures are:
        1. Go through all markdown files in the directory.
        2. For each markdown file, specify an assets directory as `<markdown_file_name>.assets` in the same directory as the markdown file.
        3. Search the file content, find and save all image blocks:
            3.1 If the block is a html image block, get the `src`, `alt` and `style` attributes.
            3.2 If the block is a markdown image block, get the `src` and `alt` attributes.
            3.3 Normalize the `alt` attribute and use it as the image file name in next steps.
            3.4 If the image is a network image, download it and save it to the assets directory.
            3.5 If the image is a local image, copy it to the assets directory, ignore its original path.  If it is already in the assets directory, it will be refreshed.
        4. Modify the image blocks to match the changes above.
        5. Delete the assets directory if it is empty.
        6. Delete the image files in the assets directory if they are not used in the markdown file.
    Args:
        directory (str): the directory that contains all target markdown files
        
    """
    paths = pathlib.Path(directory).glob('**/*.md', case_sensitive=None)
    html_img_pattern = r"<img\s*(?:src=[\"\'](?P<src>[^\"\']*)[\"\'])?\s*(?:alt=[\"\'](?P<alt>[^\"\']*)[\"\']\s*)?(?:style=[\"\'](?P<style>[^\"\']*)[\"\'])?\s*/>"
    markdown_img_pattern = r"!\[(?P<alt>[^\]]*)\]\((?P<src>[^\n\r]*)\)"
    paths:list[pathlib.Path] = list(paths)
    with alive_bar(len(paths)) as bar:
        for path in paths:

            assets_dir = path.parent / f'{path.stem}.assets'
            bar()
            with open(path, "r+", encoding="utf-8") as file:
                markdown_content = file.read()
              
                _sub_html_img_pattern = functools.partial(sub_html_img_pattern,img_save_dir=assets_dir)
                _sub_markdown_img_pattern = functools.partial(sub_markdown_img_pattern,img_save_dir=assets_dir)
                markdown_content = re.sub(html_img_pattern, _sub_html_img_pattern, markdown_content)
                markdown_content = re.sub(markdown_img_pattern, _sub_markdown_img_pattern, markdown_content)
                image_paths = []
                for item in re.findall(html_img_pattern, markdown_content):
                    image_paths.append(path.parent/item[0])
                for item in re.findall(markdown_img_pattern, markdown_content):
                    image_paths.append(path.parent/item[1])
                for item in image_paths:
                    if not item.exists():
                        logger.error(f"image not exists: {path}")
                        logger.error(f"image not exists: {item}")
                file.seek(0)
                file.truncate() # clear all contents
                file.write(markdown_content)
            if not assets_dir.exists():
                continue
            elif not any(assets_dir.iterdir()):
                assets_dir.rmdir()
                logger.warning(f"目录 '{assets_dir}' 已删除")   
            else:
                existent_image_paths = list(assets_dir.glob('*.*'))
                for item in existent_image_paths:
                    # logger.info(f"图片 '{item}'")
                    if item not in image_paths:
                        item.unlink()
                        logger.warning(f"图片 '{item}' 已删除")
            
            # if not any(assets_dir.iterdir()):
            #     print(f"目录 '{assets_dir}' 已删除")
            
if __name__ == "__main__":
    if len(sys.argv) != 2:
        print(f"Usage: python {script_file_name} <filename>")
        sys.exit(1)

    # 从命令行参数中获取文件名
    directory = sys.argv[1]
    try:
        main(directory=directory)
    except Exception as e:
        print(f"An error occurred: {str(e)}")
   


import sys 
import json

with open('hexo_footnote_test_pandoce.json', 'r', encoding='utf-8') as json_file:
    data = json.load(json_file)

with open('formatted_output.json', 'w', encoding='utf-8') as formatted_file:
    json.dump(data, formatted_file, indent=4, ensure_ascii=False)


#!/usr/bin/env python
import panflute as pf
import re
import urllib.parse
import typeguard

import logging
import logging.handlers
import subprocess
from bs4 import BeautifulSoup
import hashlib


def main(file_name:str):
    with open(file_name, 'r+', encoding='utf-8') as file:
   
        soup = BeautifulSoup(file,features="html.parser")

        # 查找所有具有 class="footnote-ref" 的元素
        # **kwargs,关键字参数可以指定属性来搜索
        # https://www.crummy.com/software/BeautifulSoup/bs4/doc.zh/#keyword
        # class是python的关键字，所以用class_
        footnote_ref_elements = soup.find_all(
            name='a',
            string=re.compile(pattern=r"[\d]+"),
            class_='footnote-ref',
            href=re.compile(pattern=r"#fn[\d]+"),
            id=re.compile(pattern=r"fnref[\d]+"),
            role='doc-noteref'
            ) # [<a class="footnote-ref" href="#fn1" id="fnref1" role="doc-noteref"><sup>1</sup></a>,...]
        footnote_ref_dict = {}
        for item in footnote_ref_elements:
            fn_name1 = re.match(r"^#(fn[\d]+)",item.attrs['href']).groups()[0] # href="fnref1"
            fn_name2 = ''.join(re.match(r"^(fn)ref([\d]+)",item.attrs['id']).groups()) # id="fnref1"
            assert fn_name1 == fn_name2
            footnote_ref_dict[fn_name1] = item
        # print(footnote_ref_dict) 
        # {fn1': <a class="footnote-ref" href="#fn1" id="fnref1" role="doc-noteref"><sup>1</sup></a>,...}
        
        
        # 查找所有具有 class="footnote-back" 的元素, 以父级<li>元素为单位
        footnote_back_elements = soup.find_all(
            name='a',
            class_='footnote-back',
            href=re.compile(pattern=r"#fnref[\d]+"),
            role='doc-backlink') # [<a class="footnote-back" href="#fnref1" role="doc-backlink">↩︎</a>,...]
        footnote_back_li_elements = [item.find_parent('li',id=re.compile(pattern=r"fn[\d]*")) for item in footnote_back_elements]
        # [<li id="fn1"><p>https://zzz<a class="footnote-back" href="#fnref1" role="doc-backlink">↩︎</a></p></li>,...]
        
        footnote_back_dict = {}
        for item in footnote_back_li_elements:
            fn_name1 = re.match(r"^(fn[\d]+)",item.attrs['id']).groups()[0] #  id="fn1"
            fn_name2 = ''.join(re.match(r"^#(fn)ref([\d]+)",item.p.a.attrs['href']).groups()) # href="#fnref1"
            assert fn_name1 == fn_name2
            footnote_back_dict[fn_name1] = item
        # print(footnote_back_dict)
        # {'fn1': <li id="fn1"><p>https://zzz<a class="footnote-back" href="#fnref1" role="doc-backlink">↩︎</a></p></li>,...}
        
        
        ##  De-duplicate the footnote-back elements:
        # Group all footnote-back elements according to the footnote-back content.
        # In each group, merge all of the footnote-back elements into one, to realize the result like:
        # <li id="fn_name">
        #   <p>content
        #       <a class="footnote-back" href="#fn_name_1" role="doc-backlink">↩︎</a>
        #       <a class="footnote-back" href="#fn_name_2" role="doc-backlink">↩︎</a>
        #       <a class="footnote-back" href="#fn_name_3" role="doc-backlink">↩︎</a>
        #   </p></li>   
        # Modify the index (content|string) of the related footnote-ref elements simultaneously.
        non_duplication_fn_dict = {}
        fn_name_group_dict = {} 
        for fn_name,item in footnote_back_dict.items():
            hash_getter = hashlib.new('md5') # there is no need to use strong hash algorithm
            hash_getter.update(item.p.contents[0].encode('utf-8'))
            hash_value = hash_getter.hexdigest()
            if hash_getter.hexdigest() not in non_duplication_fn_dict:
                non_duplication_fn_dict[hash_value] = fn_name 
                fn_name_group_dict[fn_name] = [fn_name]
                # fn_merge_vector.append((fn_name,fn_name))
            else:
                # fn_merge_vector.append((fn_name,non_duplication_fn_dict[hash_value]))
                (fn_name_group_dict[non_duplication_fn_dict[hash_value]]).append(fn_name)
        # print(non_duplication_fn_dict)
        print(fn_name_group_dict)
        
        for global_index,(fn_name,fn_group) in enumerate(fn_name_group_dict.items()):
            _global_index = global_index+1
            for group_index,single_fn_name in enumerate(fn_group):
                if fn_name == single_fn_name:
                    assert group_index == 0
                    footnote_ref_dict[single_fn_name].sup.string = str(_global_index) 
                    footnote_ref_dict[single_fn_name].attrs['href'] = f"#fn{_global_index}"
                    footnote_ref_dict[single_fn_name].attrs['id'] = f"fnref{_global_index}"
                    footnote_back_dict[single_fn_name].p.a.attrs['href'] = f"#fnref{_global_index}"
                    footnote_back_dict[single_fn_name].p.a.attrs['id'] = f"fn{_global_index}"
                    # print(footnote_back_dict[single_fn_name].p.a.attrs['id'])
                    continue
                else:
                    _append_index = group_index
                    footnote_ref_dict[single_fn_name].sup.string = str(_global_index) 
                    # print(footnote_ref_dict[single_fn_name].sup.string)
                    footnote_ref_dict[single_fn_name].attrs['href'] = f"{footnote_ref_dict[fn_name].attrs['href']}-{_append_index}"
                    # print(footnote_ref_dict[single_fn_name].attrs['href'])
                    footnote_ref_dict[single_fn_name].attrs['id'] = f"{footnote_ref_dict[fn_name].attrs['id']}-{_append_index}"
                    # print(footnote_ref_dict[single_fn_name].attrs['id'])
                
                    
                    footnote_back_dict[single_fn_name].p.a.attrs['href'] = f"{footnote_back_dict[fn_name].p.a.attrs['href']}-{_append_index}"
                    # print(footnote_back_dict[single_fn_name].p.a.attrs['href'])
                    footnote_back_dict[single_fn_name].p.a.attrs['id'] = f"{footnote_back_dict[fn_name].p.a.attrs['id']}-{_append_index}"
                    # print(footnote_back_dict[single_fn_name].p.a.attrs['id'])
                    
                    footnote_back_dict[fn_name].p.append(footnote_back_dict[single_fn_name].p.a) # move and will delete the original one
                    footnote_back_dict[single_fn_name].extract() # extract from the tree (soup)
                    del footnote_back_dict[single_fn_name].attrs['id']
                    # (even though not delete from the memory directly, it will work as deletion if regenerate a new html from the modified tree)
            del footnote_back_dict[fn_name].attrs['id']      
        # print(footnote_ref_dict)
        # print(footnote_back_dict)

        file.seek(0)
        file.truncate() # clear all contents
        file.write(str(soup))
        # print("Modified HTML saved to 'modified.html'")

if __name__ == "__main__":
    
    import sys

    # 检查命令行参数是否包含文件名
    if len(sys.argv) != 2:
        print("Usage: python test.py <filename>")
        sys.exit(1)

    # 从命令行参数中获取文件名
    filename = sys.argv[1]
    # print(filename)
    try:
        
        main(filename)
    

    except FileNotFoundError:
        print(f"File '{filename}' not found.")
    except Exception as e:
        print(f"An error occurred: {str(e)}")



import pathlib
import logging
import functools
import panflute as pf

from pandoc_filter.utils import TracingLogger
from pandoc_filter.md2html_filters import anchor_filter,internal_link_recorder,link_like_filter
from pandoc_filter.md2md_filters import internal_link_filter

pathlib.Path("./logs").mkdir(parents=True, exist_ok=True)
tracing_logger = TracingLogger(name="./logs/pf_log",level=logging.INFO)

def _check_file_path(file_path:str)->pathlib.Path:
    file_path:pathlib.Path = pathlib.Path(file_path)
    assert file_path.exists()
    assert file_path.is_file()
    return file_path

def finalize(doc:pf.Doc,**kwargs):
    tracing_logger = kwargs['tracing_logger']
    id_set = set()
    for k,v in doc.anchor_count.items():
        for i in range(1,v+1):
            id_set.add(f"{k}-{i}")
    for patched_elem,url,guessed_url_with_num in doc.internal_link_record:
        if f"{url}-1" in id_set:
            patched_elem.sub(f"{url}-1",tracing_logger)
        elif guessed_url_with_num in id_set: # None is not in id_set
            patched_elem.sub(f"{guessed_url_with_num}",tracing_logger)
        else:
            tracing_logger.logger.warning(f"{patched_elem.elem}")
            tracing_logger.logger.warning(f"The internal link `{url}` is invalid and will not be changed because no target header is found.")
    
def test_md2html_anchor_filter():
    file_path =_check_file_path("./resources/test_md2html_anchor_and_link.md")
    with open(file_path,'r',encoding='utf-8') as f:
        markdown_content = f.read()
    output_path = pathlib.Path(f"./temp/{file_path.stem}.html")
    doc = pf.convert_text(markdown_content,input_format='markdown',output_format='panflute',standalone=True)
    doc = pf.run_filter(action=internal_link_filter,doc=doc,tracing_logger=tracing_logger)
    _finalize = functools.partial(finalize,tracing_logger=tracing_logger)
    doc = pf.run_filters(actions=[anchor_filter,internal_link_recorder,link_like_filter],doc=doc,finalize=_finalize,tracing_logger=tracing_logger)
    with open(output_path, "w", encoding="utf-8") as f:
        f.write(pf.convert_text(doc,input_format='panflute',output_format='html',standalone=True))


import typeguard
import hashlib
import urllib.parse
import panflute as pf
from .utils.logging_helper import TracingLogger
from .utils import html_helper
from .md2md_filters import _decode_internal_link_url
import re

@typeguard.typechecked     
def figure_filter_v1(elem:pf.Element,doc:pf.Doc,**kwargs)->None: # Modify In Place:
    r"""Follow the general procedure of [Panflute](http://scorreia.com/software/panflute/)
    Deprecated. 
    The best way is to use CSS files to define global styles and use them by `--css <css_files>`,
    instead of make a filter here.
    So, this method is deprecated.        
    
    When converting markdown to html, Pandoc will generate a `Figure` element for each image,
    and copy the `alt_text` in `![image_url](alt_text)` to the `Caption` part of the `Figure`.
    
    And if in many Blog systems, the `alt_text` in `Image` will be copied again into the `Figure`,
    parallel with the original `Caption` part. Leading to dual captions in html.
    
    A not-so-good solution is to delete/elimiate the `Caption` part of the `Figure` element,
    which relys on the post-processing of the html file.
    
    A better way is to set inline styles on the `Figure` element, which does not rely on
    the post-processing components.
    
    Since in `Panflute`, the `Caption` element has no attributes, it is hard to modify it.
    So, we have to insert a `Div` element into the `Caption` element, and modify the `Div`
    element to show our target `alt_text`.
    
    And, it is better if set a global style `text-align:center` for the whole `Figure` element,
    which will not only influence the `Caption` part, but also the `Image` part.
    
    Even though in some Blog systems, the `Figure` element will be centered automatically,
    it is still a good practice to set the style in the `Figure` element, which will not
    be limited by post-processing components.
    
    For specific, the `Figure` element will be modified as:
        - add `text-align:center;` to the `style` attribute of the `Figure` element
        - new a `Caption` element, which includes a new `Div` element that contains the original `alt_text`
        - add `color:#858585;` to the `style` attribute of the `Div` element
        - replace the original `Caption` element with the new one.
        
    """
    tracing_logger = kwargs['tracing_logger']
    if isinstance(elem, pf.Figure):
        tracing_logger.mark(elem)
        for img in elem.content:
            if isinstance(img, pf.Image):
                break
        if 'style' in elem.attributes:
            elem.attributes['style'] += "text-align:center;"
        else:
            elem.attributes = {'style':"text-align:center;"}
        centered_div = pf.Div(*elem.caption.content,attributes={'style':"color:#858585;"})
        elem.caption = pf.Caption(centered_div)
        tracing_logger.check_and_log('figure',elem)
        
  


r"""Defination
In Markdown:
    anchors:
        headings anchors: `## aaa`
        raw-HTML anchors: `<a id="aaa"></a>`
    links:
        internal links:
            md internal links: `[bbb](#aaa)`
            raw-HTML internal links: `<a href="#aaa">bbb</a>`
        ...
    ...
"""
def __get_hash(text:str)->str:
    text = urllib.parse.unquote(text).strip(' ').lower() # 需要统一小写以获取更大的兼容性
    return hashlib.sha256(text.encode('utf-8')).hexdigest()
@typeguard.typechecked
def anchor_filter(elem:pf.Element,doc,**kwargs)->None: # Modify In Place:
    r"""Follow the general procedure of [Panflute](http://scorreia.com/software/panflute/)
    A filter to normalize anchors when converting markdown to html.
    """
    tracing_logger = kwargs['tracing_logger']
    if not(hasattr(doc,'anchor_count') and isinstance(doc.anchor_count,dict)):
        doc.anchor_count = {}
    def _text_hash_count(text:str)->str:
        text_hash = __get_hash(text)
        if text_hash in doc.anchor_count: # 按照text_hash值计数, 重复则加1
            doc.anchor_count[text_hash] += 1
        else:
            doc.anchor_count[text_hash] = 1
        return text_hash
    if isinstance(elem, pf.Header):
        tracing_logger.mark(elem)
        # 获取header文本内容并剔除#号
        header_text = pf.convert_text(elem,input_format='panflute',output_format='gfm',standalone=True).lstrip('#')
        text_hash = _text_hash_count(header_text)
        elem.identifier = f"{text_hash}-{doc.anchor_count[text_hash]}"
        tracing_logger.check_and_log('headings anchor',elem)
    elif isinstance(elem, pf.RawInline) and elem.format == 'html' and (raw_id_text:=html_helper.get_id(elem.text)): # 获取id文本内容但不做任何剔除
        tracing_logger.mark(elem)
        text_hash = _text_hash_count(raw_id_text)
        elem.text = html_helper.sub_id(elem.text,f"{text_hash}-{doc.anchor_count[text_hash]}")
        tracing_logger.check_and_log('raw-HTML anchor',elem)
        
class _PatchedInternalLink:
    @typeguard.typechecked
    def __init__(self,elem:pf.Link) -> None:
        self.elem = elem
    @typeguard.typechecked
    def sub(self,url:str,tracing_logger:TracingLogger)->None:
        tracing_logger.mark(self.elem)
        self.elem.url = f"#{url}"
        tracing_logger.check_and_log('internal_link',self.elem)

class _PatchedInternalRawLink:
    @typeguard.typechecked
    def __init__(self,elem:pf.RawInline) -> None:
        self.elem = elem
    @typeguard.typechecked
    def sub(self,url:str,tracing_logger:TracingLogger)->None:
        tracing_logger.mark(self.elem)
        self.elem.text = html_helper.sub_href(self.elem.text,f"#{url}")   
        tracing_logger.check_and_log('internal_link',self.elem)

@typeguard.typechecked
def internal_link_recorder(elem:pf.Element, doc:pf.Doc,**kwargs)->None: # Do not modify.
    r"""Follow the general procedure of [Panflute](http://scorreia.com/software/panflute/)
    A recorder to pre-normalize and record internal links when converting markdown to html.
    Since internal links should match anchors, if we want to normalize internal links better, we should consider global information comprehensively.
    So, the modification on internal links should be performed at the global level but not the node level as here.
    """
    if not(hasattr(doc,'internal_link_record') and isinstance(doc.internal_link_record,list)):
        doc.internal_link_record = []
    def _url_hash_guess(text:str)->str:
        old_url = text.lstrip('#')
        url = __get_hash(old_url)
        if match:= re.search(r'-(?P<guessed_index>\d+)$', old_url):
            guessed_index = int(match.groupdict()['guessed_index'])  # 从匹配结果中提取数字部分
            if guessed_index == 0:
                guessed_index = 1
            url_striped = old_url[:match.start()]  # 从匹配结果中提取数字前面的部分
            guessed_url_with_num = __get_hash(url_striped) + f"-{guessed_index}"
        else:
            guessed_url_with_num = None
        return url,guessed_url_with_num
      
    if isinstance(elem, pf.Link) and elem.url.startswith('#'):
        # Olny md internal links need to be decoded since it will be encoded by pandoc before filter.
        decoded_url = _decode_internal_link_url(elem.url) 
        url,guessed_url_with_num = _url_hash_guess(decoded_url)
        doc.internal_link_record.append((_PatchedInternalLink(elem),url,guessed_url_with_num))
    elif isinstance(elem, pf.RawInline) and elem.format == 'html' and (old_href:=html_helper.get_href(elem.text)) and old_href.startswith('#'):
        # raw-HTML internal links will not be encoded by pandoc before filter. So there is no need to decode it.
        url,guessed_url_with_num = _url_hash_guess(old_href)
        doc.internal_link_record.append((_PatchedInternalRawLink(elem),url,guessed_url_with_num))
        
@typeguard.typechecked
def link_like_filter(elem:pf.Element,doc:pf.Doc,**kwargs)->pf.Link|None: # Repleace
    r"""Follow the general procedure of [Panflute](http://scorreia.com/software/panflute/)
    A filter to process a string that may be like a link. Replace the string with a `Link` element.
    """
    tracing_logger = kwargs['tracing_logger']
    if isinstance(elem, pf.Str) and elem.text.lower().startswith('http'):
        tracing_logger.mark(elem)
        link = pf.Link(elem,url=elem.text)
        tracing_logger.check_and_log('link_like',elem)
        return link



import typeguard
import logging
import logging.handlers
from typing import Any
@typeguard.typechecked
def logger_factory(name:str,level:int=logging.INFO)->logging.Logger:
    r"""Generate a logger with file and console handler.
    Args:
        name: the name of the logger and the indicator of the log file
        level: the level of the logger
    Returns:
        logger: an instance of logging.Logger that has been configured
    """
    logger = logging.getLogger(name)
    logger.setLevel(level)
    # check if the console handler exists
    console_handler_exists = any(
        isinstance(handler, logging.StreamHandler)
        for handler in logger.handlers
    )
    file_handler_exists = any(
        isinstance(handler, logging.FileHandler)
        for handler in logger.handlers
    )    
    file_handler = logging.handlers.RotatingFileHandler(
        f"{name}.log",
        maxBytes=2*1024*1024,
        backupCount=5,encoding='utf-8')
    file_handler.setLevel(logging.DEBUG)
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    file_handler.setFormatter(formatter)
    console_handler.setFormatter(formatter)
    if not file_handler_exists:
        logger.addHandler(file_handler)
    if not console_handler_exists:
        logger.addHandler(console_handler)
    logger.propagate = False
    return logger

class TracingLogger:
    def __init__(self,name:str,level:int=logging.INFO) -> None:
        self.logger = logger_factory(name=name,level=level)
        self.__buf = None
    def mark(self,buf:Any):
        self.__buf = str(buf)
    def check_and_log(self,indicator:str,buf:Any):
        if self.__buf != str(buf):
            self.logger.info(f"---{indicator}:{self.__buf}")
            self.logger.info(f"+++{indicator}:{str(buf)}")